from django.db import models
# from django.conf import settings
# from django.contrib.auth.models import Group

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

import ldap3

import petname

from django.utils.module_loading import import_string

# Store LDAP info about the created groups so that we can easily
# identify them in subsequent syncs

HELP_TEXT = ('DO NOT edit this unless you really know '
             'what your doing. It is much safer to delete '
             'this entire record and let the sync command '
             'recreate it.')


class LDAPConnection(models.Model):
    """Equivalent to an ldap3.Connection."""

    user = models.CharField(max_length=255, blank=True, null=True,
                            help_text='Bind username for the LDAP server. Format is server dependent but is usually the distinguished name of the user.')
    password = models.CharField(max_length=255, blank=True, null=True,
                                help_text='Password for the bind user.')
    pool = models.ForeignKey('LDAPPool', related_name='connections')
    auto_bind = models.CharField(max_length=128,
                                 default=ldap3.AUTO_BIND_NONE,
                                 choices=[(ldap3.AUTO_BIND_NONE, 'ldap3.AUTO_BIND_NONE'),
                                          (ldap3.AUTO_BIND_NO_TLS, 'ldap3.AUTO_BIND_NO_TLS'),
                                          (ldap3.AUTO_BIND_TLS_AFTER_BIND, 'ldap3.AUTO_BIND_TLS_AFTER_BIND'),
                                          (ldap3.AUTO_BIND_TLS_BEFORE_BIND, 'ldap3.AUTO_BIND_TLS_BEFORE_BIND')])
    version = models.PositiveIntegerField(default=3)
    authentication = models.CharField(max_length=128,
                                      null=True,
                                      blank=True,
                                      choices=[(ldap3.ANONYMOUS, 'ldap3.ANONYMOUS'),
                                               (ldap3.SIMPLE, 'ldap3.SIMPLE'),
                                               (ldap3.SASL, 'ldap3.SASL'),
                                               (ldap3.NTLM, 'ldap3.NTLM')])
    # TODO: Do the other client strategies make sense here?
    client_strategy = models.CharField(max_length=128,
                                       default=ldap3.SYNC,
                                       choices=[(ldap3.SYNC, 'ldap3.SYNC'),
                                                (ldap3.RESTARTABLE, 'ldap3.RESTARTABLE')])
    auto_referrals = models.BooleanField(default=True)
    sasl_mechanism = models.CharField(max_length=128,
                                      null=True,
                                      blank=True,
                                      choices=[(ldap3.EXTERNAL, 'ldap3.EXTERNAL'),
                                               (ldap3.DIGEST_MD5, 'ldap3.DIGEST_MD5'),
                                               (ldap3.KERBEROS, 'ldap3.KERBEROS'),
                                               (ldap3.GSSAPI, 'ldap3.GSSAPI')])
    sasl_credentials = models.CharField(max_length=255, blank=True, null=True, help_text='Path to an object to use as the SASL Credential.')
    read_only = models.BooleanField(default=False)
    lazy = models.BooleanField(default=False)
    check_names = models.BooleanField(default=True)
    collect_usage = models.BooleanField(default=False)
    raise_exceptions = models.BooleanField(default=False)
    pool_name = models.CharField(max_length=255, null=True, blank=True)
    pool_size = models.PositiveIntegerField(null=True, blank=True)
    pool_lifetime = models.PositiveIntegerField(null=True, blank=True)
    fast_decoder = models.BooleanField(default=True)
    receive_timeout = models.PositiveIntegerField(null=True, blank=True)
    return_empty_attributes = models.BooleanField(default=True)
    auto_range = models.BooleanField(default=True)
    use_referral_cache = models.BooleanField(default=False)
    auto_escape = models.BooleanField(default=True)
    auto_encode = models.BooleanField(default=True)

    def hydrate(self):
        """Rehydrate this LDAPConnection into a ldap.Connection object."""
        pool = self.pool.hydrate()

        return ldap3.Connection(pool,
                                user=self.user,
                                password=self.password,
                                auto_bind=self.auto_bind,
                                version=self.version,
                                authentication=self.authentication,
                                client_strategy=self.client_strategy,
                                auto_referrals=self.auto_referrals,
                                auto_range=self.auto_range,
                                sasl_mechanism=self.sasl_mechanism,
                                sasl_credentials=self.sasl_credentials,
                                check_names=self.check_names,
                                collect_usage=self.collect_usage,
                                read_only=self.read_only,
                                lazy=self.lazy,
                                raise_exceptions=self.raise_exceptions,
                                pool_name=self.pool_name,
                                pool_size=self.pool_size,
                                pool_lifetime=self.pool_lifetime,
                                fast_decoder=self.fast_decoder,
                                receive_timeout=self.receive_timeout,
                                return_empty_attributes=self.return_empty_attributes,
                                use_referral_cache=self.use_referral_cache,
                                auto_escape=self.auto_escape,
                                auto_encode=self.auto_encode)

    def __str__(self):
        return u'LDAPConnection({})'.format(','.join([i.host for i in self.pool.servers.all()]))

    class Meta:
        verbose_name = 'ldap3.Connection'
        verbose_name_plural = 'ldap3.Connections'


class LDAPPool(models.Model):
    """Equivalent to an ldap3.ServerPool."""

    servers = models.ManyToManyField('LDAPServer', related_name='pools')
    active = models.BooleanField(default=True)
    exhaust = models.BooleanField(default=False)
    pool_strategy = models.CharField(max_length=128,
                                     default=ldap3.ROUND_ROBIN,
                                     choices=[(ldap3.FIRST, 'ldap3.FIRST'),
                                              (ldap3.ROUND_ROBIN, 'ldap3.ROUND_ROBIN'),
                                              (ldap3.RANDOM, 'ldap3.RANDOM')])

    def hydrate(self):
        """Rehydrate this LDAPPool into an ldap3.ServerPool object."""
        hydrated_servers = [s.hydrate() for s in self.servers.all()]

        return ldap3.ServerPool(servers=hydrated_servers,
                                pool_strategy=self.pool_strategy,
                                active=self.active,
                                exhaust=self.exhaust)

    def __str__(self):
        return u'LDAPPool({})'.format(','.join([i.host for i in self.servers.all()]))

    class Meta:
        verbose_name = 'ldap3.ServerPool'
        verbose_name_plural = 'ldap3.ServerPools'


class LDAPServer(models.Model):
    """Equivalent to an ldap3.Server."""

    host = models.CharField(max_length=255)
    port = models.PositiveIntegerField(blank=True, null=True)
    use_ssl = models.BooleanField(default=False)
    get_info = models.CharField(max_length=128,
                                default=ldap3.SCHEMA,
                                choices=[(ldap3.NONE, 'ldap3.NONE'),
                                         (ldap3.DSA, 'ldap3.DSA'),
                                         (ldap3.SCHEMA, 'ldap3.SCHEMA'),
                                         (ldap3.ALL, 'ldap3.ALL'),
                                         (ldap3.OFFLINE_EDIR_8_8_8, 'ldap3.OFFLINE_EDIR_8_8_8'),
                                         (ldap3.OFFLINE_AD_2012_R2, 'ldap3.OFFLINE_AD_2012_R2'),
                                         (ldap3.OFFLINE_SLAPD_2_4, 'ldap3.OFFLINE_SLAPD_2_4'),
                                         (ldap3.OFFLINE_DS389_1_3_3, 'ldap3.OFFLINE_DS389_1_3_3')])
    mode = models.CharField(max_length=128,
                            default=ldap3.IP_SYSTEM_DEFAULT,
                            choices=[(ldap3.IP_SYSTEM_DEFAULT, 'ldap3.IP_SYSTEM_DEFAULT'),
                                     (ldap3.IP_V4_ONLY, 'ldap3.IP_V4_ONLY'),
                                     (ldap3.IP_V6_ONLY, 'ldap3.IP_V6_ONLY'),
                                     (ldap3.IP_V4_PREFERRED, 'ldap3.IP_V4_PREFERRED'),
                                     (ldap3.IP_V6_PREFERRED, 'ldap3.IP_V6_PREFERRED')])
    tls = models.CharField(max_length=255, blank=True, null=True, help_text='Path to a python object which contains TLS certificate information. LEAVE THIS BLANK.')
    connect_timeout = models.PositiveIntegerField(null=True, blank=True)
    allowed_referral_hosts = models.ManyToManyField('LDAPReferralHost', blank=True)

    def hydrate(self):
        """Rehydrate this LDAPServer into a ldap3.Server object."""
        # TODO: Think of a way to include a custom formatter.
        return ldap3.Server(self.host,
                            port=self.port,
                            use_ssl=self.use_ssl,
                            allowed_referral_hosts=[(h.hostname, h.allowed) for h in self.allowed_referral_hosts.all()],
                            get_info=self.get_info,
                            tls=self.tls,
                            connect_timeout=self.connect_timeout,
                            mode=self.mode)

    def __str__(self):
        if self.port is not None:
            return u'LDAPServer({}:{})'.format(self.host, self.port)
        else:
            return u'LDAPServer({})'.format(self.host)

    class Meta:
        verbose_name = 'ldap3.Server'
        verbose_name_plural = 'ldap3.Servers'


class LDAPReferralHost(models.Model):
    """ldap3.ReferralHost."""

    hostname = models.CharField(max_length=255)
    allowed = models.BooleanField(default=True)

    def __str__(self):
        """Unicode string representation of this Referral Host."""
        return u'LDAPReferralHost(hostname={})'.format(self.hostname)

    class Meta:
        verbose_name = 'ldap3.ReferralHost'
        verbose_name_plural = 'ldap3.ReferralHosts'


class LDAPSyncJob(models.Model):
    """Configuration for a Synchronisation Job."""

    name = models.CharField(max_length=255, default=petname.Generate(3, '.'), unique=True)
    target_django_model = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    synchronizer_class = models.CharField(max_length=255, default='ldap3_sync.utils.PETLSynchronizer')
    # sync_runner_class = models.CharField(max_length=255,
    #                                      default='ldap3_sync.utils.ScheduledSyncRunner',
    #                                      help_text='dotted path to the sync runner class.')
    ldap_connection = models.ForeignKey(LDAPConnection)
    base_ou = models.TextField()  # DN's can be quite long.
    search_filter = models.TextField()
    search_scope = models.CharField(max_length=100,
                                    default=ldap3.SUBTREE,
                                    choices=[(ldap3.BASE, 'ldap3.BASE'),
                                             (ldap3.LEVEL, 'ldap3.LEVEL'),
                                             (ldap3.SUBTREE, 'ldap3.SUBTREE')])
    ldap_key_attribute = models.CharField(max_length=255, help_text='The name of a unique attribute in the LDAP data that will be used for syncing.')
    django_key_attribute = models.CharField(max_length=255, help_text='The name of a unique attribute on the Django model that will be used for syncing.')
    do_create = models.BooleanField(default=True, help_text='If True then new models will ')
    do_update = models.BooleanField(default=True, help_text='U')
    do_delete = models.BooleanField(default=True)
    delete_action = models.CharField(max_length=50,
                                     default='DELETE',
                                     choices=(('DELETE', 'DELETE'), ('FUNCTION', 'Function')))
    delete_function = models.CharField(max_length=255, blank=True)  # String that points to a function that can be used for removal.
    logging_level = models.CharField(max_length=50, default='INFO', choices=[('DEBUG', 'DEBUG'),
                                                                             ('INFO', 'INFO'),
                                                                             ('WARNING', 'WARNING'),
                                                                             ('ERROR', 'ERROR'),
                                                                             ('CRITICAL', 'CRITICAL')])

    def __str__(self):
        """Return unicode string for this LDAPSync model."""
        return u'LDAP Sync Job: {}'.format(self.name)

    class Meta:
        verbose_name = 'Synchronizer Job'
        verbose_name_plural = 'Synchronizer Jobs'


class LDAPSyncAttributeMap(models.Model):
    """Maps LDAP attributes to Model Attributes."""

    sync_job = models.ForeignKey('LDAPSyncJob', related_name='attributes')
    ldap_attribute_name = models.CharField(max_length=255, help_text="LDAP Attribute Name.")
    model_attribute_name = models.CharField(max_length=255, help_text="Django Attribute Name.")

    class Meta:
        verbose_name = 'Synchronizer Attribute Map'
        verbose_name_plural = 'Synchronizer Attribute Maps'


class LDAPSyncDjangoQuerysetFilter(models.Model):
    """A filter that will be used in the django search to reitreve matching model objects."""

    sync_job = models.ForeignKey('LDAPSyncJob', related_name='filters')
    predicate = models.CharField(max_length=255, help_text='Filter predicate.')
    value = models.CharField(max_length=255, help_text='Value to filter against.')

    class Meta:
        verbose_name = 'Django Queryset Filter'
        verbose_name = 'Django Queryset Filters'


class LDAPSyncJobLog(models.Model):
    """The logging output from a LDAPSyncJob run."""

    sync_job = models.ForeignKey('LDAPSyncJob', related_name='logs')
    log_data = models.TextField(blank=True)
    started_at = models.DateTimeField()
    ended_at = models.DateTimeField()
    successful = models.BooleanField(help_text='Simple boolean indicator as to wether this Synchonization job was successful.')

    @classmethod
    def store_log(cls, sync_job, log_data, started_at, ended_at, successful):
        """Convenience class method to take a sync_job and some log_data and store it. Will return the saved LDAPSyncJobLog."""
        l = LDAPSyncJobLog(sync_job=sync_job,
                           log_data=log_data,
                           started_at=started_at,
                           ended_at=ended_at,
                           successful=successful)
        return l.save()

    class Meta:
        verbose_name = 'Synchronizer Log'
        verbose_name_plural = 'Synchronizer Logs'
        ordering = ['started_at']



__version__ = '1.0.1'
# MAJOR version when you make incompatible API changes,
# MINOR version when you add functionality in a backwards-compatible manner, and
# PATCH version when you make backwards-compatible bug fixes.



"""django-ldap3-sync utility classes including the main synchornizer."""
import petl
from petl_ldap3 import fromldap
from petl_django import fromdjango
import logging
from io import StringIO
from django.utils.module_loading import import_string
from datetime import datetime
from .models import LDAPSyncJob, LDAPSyncJobLog


class Synchronizer(object):
    """A base synchornizer class."""
    def __init__(self, sync_job, logger):
        self.sync_job = sync_job
        self.logger = logger

    def setup(self):
        pass

    def synchronize(self):
        raise NotImplemented

    def teardown(self):
        pass

    def build_queryset(self):
        """Build a queryset using the target_django_model and LdapSyncQuerysetFilter data."""
        model_class = self.sync_job.target_django_model.model_class()
        filters = [(i.predicate, i.value) for i in self.sync_job.filters.all()]
        if len(filters) > 0:
            return model_class.objects.filter(**dict(filters)).all()
        else:
            return model_class.objects.all()

    def build_attribute_map_for_ldap(self):
        """Build a dictionary of attribute name pairs mapping from django to ldap names."""
        return dict([(i.model_attribute_name, i.ldap_attribute_name) for i in self.sync_job.attributes.all()])

    def build_attribute_map_for_django(self):
        """Build a dictionary of attribute name pairs mapping from ldap to django names."""
        return dict([(i.ldap_attribute_name, i.model_attribute_name) for i in self.sync_job.attributes.all()])


class PETLSynchronizer(Synchronizer):
    """A Synchronizer class that uses the PETL library for data manipulation and takes an LDAPSyncJob as a configuration item."""

    @property
    def data_for_create(self):
        if not hasattr(self, '_data_for_create'):
            self._data_for_create = petl.antijoin(self.ldap_data, self.django_data,
                                                  lkey=self.sync_job.ldap_key_attribute,
                                                  rkey=self.sync_job.django_key_attribute)
        return self._data_for_create

    @property
    def data_for_update(self):
        if not hasattr(self, '_data_for_update'):
            self._data_for_update = petl.antijoin(self.ldap_data, self.data_for_create,
                                                  key=self.sync_job.ldap_key_attribute)
        return self._data_for_update

    @property
    def data_for_delete(self):
        if not hasattr(self, '_data_for_delete'):
            self._data_for_delete = petl.antijoin(self.django_data, self.ldap_data,
                                                  lkey=self.sync_job.django_key_attribute,
                                                  rkey=self.sync_job.ldap_key_attribute)
        return self._data_for_delete

    def creation_stage(self):
        """The Creation Stage of the synchonization process."""
        self.logger.debug('Begin generating data for new model creation.')
        try:
            data_for_create = self.data_for_create
        except BaseException as e:
            self.logger.exception('Unable to generate data for new model creation.')
            raise e
        else:
            self.logger.info('Generated data for new model creation.')
            self.logger.debug('{} new models need to be created.'.format(data_for_create.nrows()))

        try:
            self.logger.debug('Renaming LDAP attributes based on the defined attribute map to match the Django attributes.')
            data_for_create = data_for_create.rename(self.build_attribute_map_for_django())
        except BaseException as e:
            self.logger.exception('Unable to rename new model data.')
            raise e

        self.logger.info('Beginning to create new models.')
        new_models_created = 0
        new_models_failed = 0
        for i in data_for_create.dicts():
            # If I do this one at a time it is slower but lets me catch errors etc.
            try:
                self.logger.info('Attempting to create new {}.'.format(self.model_class))
                self.logger.debug('Using model data: {}'.format(str(i)))
                new_model = self.model_class(**i)
                new_model.save()
            except BaseException:
                self.logger.exception('Unable to create new {}.'.format(self.model_class))
                new_models_failed += 1
            else:
                self.logger.debug('Successfully created new {}.'.format(self.model_class))
                new_models_created += 1
        self.logger.debug('Successfully created {} new models.'.format(new_models_created))
        self.logger.debug('Failed creating {} new models.'.format(new_models_failed))

    def update_stage(self):
        """Update Stage of the Synchonization process."""
        self.logger.debug('Begin generating data for existing model updates.')
        try:
            # generate updates. This is the data that exists in both stores and needs to be evaluated for update elligibility.
            # update data should be any ldap data that is not in create. Delete entries will not exist in ldap data.
            data_for_update = self.data_for_update
        except BaseException as e:
            self.logger.exception('Unable to generate data for updates.')
            raise e
        else:
            self.logger.info('Generated data for existing model updates.')
            self.logger.debug('{} candidate models for update.')

        try:
            self.logger.debug('Renaming LDAP attributes based on the defined attribute map to match the Django attributes.')
            data_for_update = data_for_update.rename(self.build_attribute_map_for_django())
        except BaseException as e:
            self.logger.exception('Unable to rename existing model update data.')
            raise e

        self.logger.info('Beginning to evaluate models for update.')
        updates_applied = 0
        updates_skipped = 0
        updates_failed = 0

        header = data_for_update.header()
        key_attribute = self.sync_job.django_key_attribute

        for i in data_for_update.dicts():
            try:
                key_attribute_value = i[key_attribute]
                existing_model = self.model_class.objects.get(**{key_attribute: key_attribute_value})
                has_updates = False
                for h in header:
                    if i[h] != getattr(existing_model, h, None):
                        setattr(existing_model, h, i[h])
                        has_updates = True
                if has_updates is True:
                    existing_model.save()
                    updates_applied += 1
                    self.logger.debug('Successfully applied updates to model with key attribute value: {}'.format(key_attribute_value))
                else:
                    updates_skipped += 1
                    self.logger.debug('Skipped applying updates to model with key attribute value: {}'.format(key_attribute_value))
            except BaseException:
                self.logger.exception('Unable to apply updates to model with key attribute value: {}'.format(key_attribute_value))
                updates_failed += 1
        self.logger.debug('Successfully applied {} updates.'.format(updates_applied))
        self.logger.debug('Skipped applying {} updates.'.format(updates_skipped))
        self.logger.debug('Failed to apply {} updates.'.format(updates_failed))

    def delete_stage(self):
        """Delete Stage of the Synchonization process."""
        self.logger.debug('Begin generating data for deletion candidates.')

        data_for_delete = self.data_for_delete
        key_attribute = self.sync_job.django_key_attribute

        if self.sync_job.delete_action == 'DELETE':
            delete_action = lambda r: r.delete()  # noqa
        elif self.sync_job.delete_action == 'FUNCTION':
            try:
                delete_action = import_string(self.sync_job.delete_function)
            except ImportError:
                self.logger.exception('Unable to import the configured delete function: {}'.format(self.sync_job.delete_function))

        successful_deletes = 0
        failed_deletes = 0

        for i in data_for_delete.dicts():
            key_attribute_value = i[key_attribute]
            existing_model = self.model_class.objects.get(**{key_attribute: key_attribute_value})
            try:
                delete_action(existing_model)
            except BaseException:
                self.logger.exception('Unable to delete model with key attribute value: {}'.format(key_attribute_value))
                failed_deletes += 1
            else:
                self.logger.debug('Successfully deleted model with key attribute value: {}'.format(key_attribute_value))
                successful_deletes += 1
        self.logger.debug('Successfully deleted {} models.'.format(successful_deletes))
        self.logger.debug('Failed to delete {} models.'.format(failed_deletes))

    def synchronize(self):
        """Perform a synchronization of ldap objects to django models as defined in a LDAPSyncJob."""
        self.logger.info('Begin Synchronization for LDAPSyncJob: {}'.format(self.sync_job.name))

        # Rehydrate the ldap_connection object from the stored configuration
        try:
            self.logger.debug('Rehydrating ldap3.Connection')
            ldap_connection = self.sync_job.ldap_connection.hydrate()
        except BaseException:
            self.logger.exception('Unable to rehydrate the ldap3.Connection.')
            return
        else:
            self.logger.info('Successfully rehydrated ldap3.Connection')
            self.logger.debug('Rehydrated ldap3.Connection: {}'.format(str(ldap_connection)))

        ldap_attributes = set([a.ldap_attribute_name for a in self.sync_job.attributes.all()])
        self.logger.debug('LDAP Attributes to be retrieved: {}'.format(','.join(ldap_attributes)))

        # Search for LDAP objects and return
        try:
            self.logger.debug('Starting to retrieve LDAP entries using ldap3.Connection: {}'.format(str(ldap_connection)))
            self.ldap_data = fromldap(ldap_connection,
                                      self.sync_job.base_ou,
                                      self.sync_job.search_filter,
                                      attributes=ldap_attributes,
                                      scope=self.sync_job.search_scope)
        except BaseException:
            self.logger.exception('Unable to retrieve LDAP entries.')
            return
        else:
            self.logger.info('Successfully Retrieved LDAP data')
            self.logger.debug('Retrieved {} LDAP entries.'.format(self.ldap_data.nrows()))
            self.logger.debug('LDAP headers retrieved: {}'.format(str(self.ldap_data.header())))

        django_attributes = set([i.model_attribute_name for i in self.sync_job.attributes.all()])
        self.logger.debug('Django model attributes to be retrieved: {}'.format(','.join(django_attributes)))

        # collect all of the django models matching the search.
        try:
            self.logger.debug('Starting to retrieve Django models.')
            self.model_class = self.sync_job.target_django_model.model_class()
            self.logger.debug('Using Django Model: {}'.format(self.model_class))
            qs = self.build_queryset()
            self.logger.debug('Using QuerySet: {}'.format(str(qs.query)))

            self.django_data = fromdjango(self.model_class, qs, fields=django_attributes)
        except BaseException:
            self.logger.exception('Unable to retrieve Django models.')
            return
        else:
            self.logger.info('Successfully retrieved Django models.')
            self.logger.debug('Retrieved {} Django models.'.format(self.django_data.nrows()))
            self.logger.debug('Django headers retrieved: {}'.format(str(self.django_data.header())))

        # ==== Create New Models ====
        if self.sync_job.do_create is True:
            try:
                self.creation_stage()
            except BaseException:
                self.logger.exception('Creation Stage Failed. Badly.')
            else:
                self.logger.debug('Creation Stage Finished Succesfully.')
        else:
            self.logger.info('Skipping the Creation Stage.')

        # ==== Update Existing Models ====
        if self.sync_job.do_update is True:
            try:
                self.update_stage()
            except BaseException:
                self.logger.exception('Update Stage Failed. Badly.')
            else:
                self.logger.debug('Update Stage Finished Succesfully.')
        else:
            self.logger.info('Skipping the Update Stage.')

        # ==== Delete Models that do not exist in the LDAP Data ====
        if self.sync_job.do_delete is True:
            try:
                self.delete_stage()
            except BaseException:
                self.logger.exception('Delete Stage Failed. Badly.')
            else:
                self.logger.debug('Delete Stage Finished Succesfully.')
        else:
            self.logger.info('Skipping the Delete Stage.')


class SyncRunner(object):
    """Base abstract sync runner class."""
    def __init__(self, sync_job):
        self.sync_job = sync_job

    def run(self):
        self.logger = self.setup_logging()

        synchronizer_class = import_string(self.sync_job.synchronizer_class)
        synchronizer_instance = synchronizer_class(self.sync_job, self.logger)

        synchronizer_instance.setup()
        synchronizer_instance.synchronize()
        synchronizer_instance.teardown()

        self.teardown_logging()

    def setup_logging(self):
        raise NotImplemented

    def teardown_logging(self):
        raise NotImplemented


class CLISyncRunner(SyncRunner):
    """CLI Runner class. Takes a name instead of a LDAPSyncJob model."""
    def setup_logging(self):
        """Setup a logger that sends everything to the console."""
        logger = logging.getLogger('LDAPSyncJob: {}'.format(self.sync_job.name))
        logger.setLevel(self.sync_job.logging_level)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(self.sync_job.logging_level)

        formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:: %(message)s')
        console_handler.setFormatter(formatter)

        logger.addHandler(console_handler)

        return logger

    def teardown_logging(self):
        """Perform any teardown actions for the logging system."""
        pass


class BackgroundSyncRunner(SyncRunner):
    """Background runner class."""
    def setup_logging(self):
        """log everything to a variable and store the contents in a LDAPSyncJobLog."""
        self.start = datetime.now()
        logger = logging.getLogger('LDAPSyncJob: {}'.format(self.sync_job.name))
        logger.setLevel(self.sync_job.logging_level)

        self.logging_target = StringIO.StringIO()
        stream_handler = logging.StreamHandler(self.logging_target)
        stream_handler.setLevel(self.sync_job.logging_level)

        formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:: %(message)s')
        stream_handler.setFormatter(formatter)

        logger.addHandler(stream_handler)

        return logger

    def teardown_logging(self):
        end = datetime.now()

        logging_data = self.logging_target.getvalue()
        self.logging_target.close()

        successful = getattr(self, 'successful', False)

        LDAPSyncJobLog.store_log(self.sync_job, logging_data, self.start, end, successful)


from __future__ import absolute_import, unicode_literals
from celery import shared_task
from .models import LDAPSyncJob
from .utils import BackgroundSyncRunner


@shared_task
def syncldap(sync_job_name):
    """Retrieve the given sync_job by name and attempt to run it."""
    try:
        sync_job = LDAPSyncJob.objects.get(sync_job_name)
    except LDAPSyncJob.DoesNotExist as e:
        return f'FAILED: {e}'
    runner = BackgroundSyncRunner(sync_job)
    runner.run()
    return 'RUN'


#!/usr/bin/env python
import os
import sys

from coverage import Coverage

import django
from django.conf import settings
from django.test.utils import get_runner


BASE_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(BASE_PATH, '../'))

if __name__ == "__main__":
    os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_settings'
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    cov = Coverage(config_file=os.path.join(BASE_PATH, 'tests/.coveragerc'),
                   source=[BASE_PATH],
                   omit=[os.path.join(BASE_PATH, 'tests/*')])
    cov.start()
    failures = test_runner.run_tests(["tests"])
    cov.stop()
    cov.html_report(directory=os.path.join(BASE_PATH, 'tests/coverage_report/'))
    sys.exit(bool(failures))



from django.core.management.base import BaseCommand, CommandError
from ldap3_sync.utils import CLISyncRunner
from ldap3_sync.models import LDAPSyncJob


class Command(BaseCommand):
    help = "Run a preconfigured LDAPSyncJob."

    def add_arguments(self, parser):
        parser.add_argument('job_name', type=str)

    def handle(self, *args, **options):
        try:
            sync_job = LDAPSyncJob.objects.get(name=options['job_name'])
        except LDAPSyncJob.DoesNotExist:
            raise CommandError('LDAPSyncJob with name {} does not exist.'.format(options['job_name']))
        runner = CLISyncRunner(sync_job)
        runner.run()



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from ldap3_sync import __version__ as version

long_description = 'Please see the documentation at the `project page <https://github.com/PGower/django-ldap3-sync>`_ .'

packages = [
    'ldap3_sync',
    'ldap3_sync.management',
    'ldap3_sync.management.commands',
]

package_data = {
    '': ['LICENSE', 'README.md'],
}

# with open('README.md') as f:
#     readme = f.read()

setup(
    name='django-ldap3-sync',
    version=version,
    description='A Django application for synchronizing LDAP objects to Django models.',
    long_description=long_description,
    author='Paul Gower',
    author_email='p.gower@gmail.com',
    url='https://github.com/PGower/django-ldap3-sync',
    download_url='https://github.com/PGower/django-ldap3-sync/releases',
    package_dir={'ldap3-sync': 'ldap3-sync'},
    packages=packages,
    package_data=package_data,
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=['django', 'ldap', 'active directory', 'synchronize', 'sync'],
    install_requires=['ldap3 >= 0.9.7.4', 'petname >= 1.12', 'petl', 'petl_ldap3', 'petl_django'],
)



from django.contrib import admin
from django.utils.module_loading import import_string

from ldap3_sync.models import (LDAPConnection,
                               LDAPPool,
                               LDAPServer,
                               LDAPReferralHost,
                               LDAPSyncJob,
                               LDAPSyncAttributeMap,
                               LDAPSyncDjangoQuerysetFilter,
                               LDAPSyncJobLog)


class LDAPConnectionAdmin(admin.ModelAdmin):
    class Meta:
        model = LDAPConnection


class LDAPPoolAdmin(admin.ModelAdmin):
    class Meta:
        model = LDAPPool


class LDAPServerAdmin(admin.ModelAdmin):
    class Meta:
        model = LDAPServer


class LDAPReferralHostAdmin(admin.ModelAdmin):
    class Meta:
        model = LDAPReferralHost


class LDAPSyncAttributeMapAdmin(admin.TabularInline):
    model = LDAPSyncAttributeMap


class LDAPSyncDjangoQuerysetFilterAdmin(admin.TabularInline):
    model = LDAPSyncDjangoQuerysetFilter


class LDAPSyncJobAdmin(admin.ModelAdmin):
    inlines = [LDAPSyncAttributeMapAdmin, LDAPSyncDjangoQuerysetFilterAdmin]
    list_display = ['name', 'get_django_model_name', 'get_synchronizer_class_name', 'get_ldap_servers', 'do_create', 'do_update', 'do_delete']

    def get_django_model_name(self, obj):
        c = obj.target_django_model.model_class()
        return c.__name__
    get_django_model_name.short_description = 'Django Model'

    def get_synchronizer_class_name(self, obj):
        c = import_string(obj.synchronizer_class)
        return c.__name__
    get_synchronizer_class_name.short_description = 'Synchronizer Class'

    def get_ldap_servers(self, obj):
        return ','.join(i.host for i in obj.ldap_connection.pool.servers.all())
    get_ldap_servers.short_description = 'LDAP Servers'


    class Meta:
        model = LDAPSyncJob


class LDAPSyncJobLogAdmin(admin.ModelAdmin):
    list_display = ['sync_job', 'started_at', 'ended_at', 'successful']
    class Meta:
        model = LDAPSyncJobLog


admin.site.register(LDAPConnection, LDAPConnectionAdmin)
admin.site.register(LDAPPool, LDAPPoolAdmin)
admin.site.register(LDAPServer, LDAPServerAdmin)
admin.site.register(LDAPReferralHost, LDAPReferralHostAdmin)
admin.site.register(LDAPSyncJob, LDAPSyncJobAdmin)
admin.site.register(LDAPSyncJobLog, LDAPSyncJobLogAdmin)
